from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Literal

from yarl import URL

from graia.amnesia.transport.interface import ExtraContent


@dataclass
class HttpRequest(ExtraContent):
    headers: dict[str, str]
    cookies: dict[str, str]
    query_params: dict[str, str]
    url: URL
    host: str
    method: Literal["GET", "POST", "PUT", "DELETE"] | str
    client_ip: str
    client_port: int


@dataclass
class HttpResponse(ExtraContent):
    status: int
    headers: dict[str, str]
    cookies: dict[str, str]
    uri: URL



from __future__ import annotations

from collections.abc import Callable, Coroutine
from typing import Any

from graia.amnesia.transport.common.status import ConnectionStatus
from graia.amnesia.transport.common.websocket.io import AbstractWebsocketIO
from graia.amnesia.transport.signature import TransportSignature

WebsocketConnectEvent: TransportSignature[
    Callable[[AbstractWebsocketIO], Coroutine[None, None, Any]]
] = TransportSignature()
WebsocketReceivedEvent: TransportSignature[
    Callable[[AbstractWebsocketIO, bytes | str], Coroutine[None, None, Any]]
] = TransportSignature()
WebsocketCloseEvent: TransportSignature[
    Callable[[AbstractWebsocketIO], Coroutine[None, None, Any]]
] = TransportSignature()
WebsocketReconnect: TransportSignature[Callable[[ConnectionStatus], Coroutine[None, None, bool]]] = TransportSignature()



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from collections.abc import Awaitable, Callable, MutableMapping
from typing import Any

from launart import ExportInterface
from launart.service import Service


class AbstractAsgiService(Service):
    supported_interface_types = set()
    host: str
    port: int

    def get_interface(self, _):
        return

    def __init__(self, host: str = "127.0.0.1", port: int = 8000):
        super().__init__()
        self.host = host
        self.port = port


class ASGIHandlerProvider(ExportInterface, metaclass=ABCMeta):
    @abstractmethod
    def get_asgi_handler(
        self,
    ) -> Callable[
        [
            type[MutableMapping[str, Any]],
            Awaitable[MutableMapping[str, Any]],
            Callable[[MutableMapping[str, Any]], Awaitable[None]],
        ],
        None,
    ]:
        ...


class ExternalASGIHandlerService(Service, ASGIHandlerProvider):
    supported_interface_types = {ASGIHandlerProvider}

    asgi: Callable[
        [
            type[MutableMapping[str, Any]],
            Awaitable[MutableMapping[str, Any]],
            Callable[[MutableMapping[str, Any]], Awaitable[None]],
        ],
        None,
    ]

    def __init__(
        self,
        asgi_handler: Callable[
            [
                type[MutableMapping[str, Any]],
                Awaitable[MutableMapping[str, Any]],
                Callable[[MutableMapping[str, Any]], Awaitable[None]],
            ],
            None,
        ],
    ) -> None:
        self.asgi = asgi_handler

    def get_interface(self, _):
        return self

    def get_asgi_handler(
        self,
    ) -> Callable[
        [
            type[MutableMapping[str, Any]],
            Awaitable[MutableMapping[str, Any]],
            Callable[[MutableMapping[str, Any]], Awaitable[None]],
        ],
        None,
    ]:
        return self.asgi



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from collections.abc import Callable
from typing import Any, TypeVar

from launart.service import ExportInterface, Service

from graia.amnesia.transport import Transport
from graia.amnesia.transport.common.http.io import AbstractServerRequestIO
from graia.amnesia.transport.common.websocket.io import AbstractWebsocketIO
from graia.amnesia.transport.rider import TransportRider

T = TypeVar("T", bound="AbstractServerService")
K = TypeVar("K")
V = TypeVar("V", bound="AbstractServerRequestIO | AbstractWebsocketIO")


class AbstractRouter(
    ExportInterface[T],
    TransportRider[K, V],
    metaclass=ABCMeta,
):
    @abstractmethod
    async def handle_http_request(self, handler: Callable[[AbstractServerRequestIO], Any], request: Any) -> Any:
        raise NotImplementedError

    @abstractmethod
    async def handle_websocket_request(self, request: Any) -> Any:
        raise NotImplementedError

    @abstractmethod
    def use(self, transport: Transport):
        raise NotImplementedError

    def io(self, id: K):
        return self.connections.get(id)


class AbstractServerService(Service):
    supported_interface_types = {AbstractRouter}



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import Any, TypeVar

from launart.service import ExportInterface, Service

from graia.amnesia.json import TJson
from graia.amnesia.transport.rider import TransportRider

T = TypeVar("T", bound="AbstractClientService")


class AbstractClientInterface(ExportInterface[T], metaclass=ABCMeta):
    @abstractmethod
    def request(
        self,
        method: str,
        url: str,
        params: dict | None = None,
        data: Any = None,
        headers: dict | None = None,
        cookies: dict | None = None,
        timeout: float | None = None,
        *,
        json: TJson | None = None,
        **kwargs: Any,
    ) -> TransportRider:
        raise NotImplementedError

    @abstractmethod
    def websocket(self, url: str, **kwargs) -> TransportRider:
        raise NotImplementedError


class AbstractClientService(Service, metaclass=ABCMeta):
    supported_interface_types = {AbstractClientInterface}



from __future__ import annotations

import json
from functools import partial
from typing import Any

from .. import JSONBackend, TJson, TJsonCustomSerializer
from ..serializers import SERIALIZER_DEFAULT, SERIALIZERS


class StdBackend(JSONBackend):
    def serialize(self, value: Any, *, custom_serializers: dict[type, TJsonCustomSerializer] | None = None) -> str:
        return json.dumps(
            value,
            default=partial(
                SERIALIZER_DEFAULT, d=custom_serializers and dict(SERIALIZERS, **custom_serializers) or SERIALIZERS
            ),
        )

    def deserialize(self, value: str) -> TJson:
        return json.loads(value)


BACKEND_INSTANCE = StdBackend()



from __future__ import annotations

import asyncio
from abc import ABCMeta, abstractmethod
from collections.abc import MutableMapping
from itertools import chain
from typing import Generic, TypeVar

from loguru import logger

from graia.amnesia.transport import Transport

K = TypeVar("K")
V = TypeVar("V")


class TransportRider(Generic[K, V], metaclass=ABCMeta):
    connections: MutableMapping[K, V]
    transports: list[Transport]

    @abstractmethod
    def io(self, id: K | None = None):
        raise NotImplementedError

    @abstractmethod
    def use(self, transport: Transport):
        raise NotImplementedError

    async def trigger_callbacks(self, event, *args, **kwargs):
        result = await asyncio.gather(
            *(c(*args, **kwargs) for c in chain.from_iterable(t.get_callbacks(event) for t in self.transports)),
            return_exceptions=True,
        )
        base_excs: set[BaseException] = set(filter(lambda e: isinstance(e, BaseException), result))
        excs: set[Exception] = set(filter(lambda e: isinstance(e, Exception), result))
        for exc in excs:
            logger.opt(exception=exc).error(exc)
        if base_excs := base_excs - excs:
            raise base_excs.pop()
